import cv2


def to_single_channel(img, conversion="gray"):
    """
    Converts the image to a single channel.

    :param img: the image to convert
    :param conversion: how to convert the BGR image (gray/r/g/b)
    :type conversion: str
    :return: the single channel image
    """
    if conversion == "gray":
        return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    elif conversion == "b":
        return img[:, :, 0]
    elif conversion == "g":
        return img[:, :, 1]
    elif conversion == "r":
        return img[:, :, 2]
    else:
        raise Exception("Unhandled conversion: %s" % conversion)


def diff_img(img1, img2):
    """
    Computes the absolute difference between two images.

    :param img1: the first image
    :param img2: the second image
    """
    return cv2.absdiff(img1, img2)


def to_bw(img, threshold):
    """
    Turns the gray image into binary.

    :param img: the image to convert
    :param threshold: the threshold to use
    :type threshold: int
    :return: the binary image
    """
    thresh, binary = cv2.threshold(img, threshold, 255, cv2.THRESH_BINARY)
    return binary


def count_diff(img):
    """
    Counts the non-zero pixels in the image.

    :param img: the image to process
    :return: the count
    :rtype: int
    """
    return cv2.countNonZero(img)


def detect_change(img1, img2, conversion, bw_threshold, change_threshold):
    """
    Returns true if there was change detected between the two images (turns them into gray images first).

    :param img1: the first image
    :param img2: the second image
    :param bw_threshold: the black/white threshold (0-255)
    :type bw_threshold: int
    :param change_threshold: the threshold for changes (0-1)
    :type change_threshold: float
    :return: the detected ratio, whether change was detected
    :rtype threshold: (float, bool)
    """
    size = img1.size
    img1 = to_single_channel(img1, conversion)
    img2 = to_single_channel(img2, conversion)
    count = count_diff(to_bw(diff_img(img1, img2), bw_threshold))
    ratio = float(count) / float(size)
    return ratio, ratio > change_threshold


import cv2
import io
import numpy as np

from wai.annotations.core.component import SinkComponent
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance

from wai.common.cli.options import TypedOption


class VideoFileWriter(
    SinkComponent[ImageObjectDetectionInstance]
):

    output_file: str = TypedOption(
        "-o", "--output",
        type=str,
        default="",
        help="the MJPG video file to write to"
    )

    fps: int = TypedOption(
        "-f", "--fps",
        type=int,
        default=25,
        help="the frames per second to use"
    )

    def consume_element(self, element: ImageObjectDetectionInstance):
        """
        Consumes instances.
        """

        img_array = np.fromstring(io.BytesIO(element.data.data).read(), dtype=np.uint8)
        img = cv2.imdecode(img_array, cv2.IMREAD_COLOR)

        # open video file
        if not hasattr(self, "_out"):
            h, w, _ = img.shape
            self._out = cv2.VideoWriter(self.output_file, cv2.VideoWriter_fourcc('M', 'J', 'P', 'G'), self.fps, (w, h))

        self._out.write(img)

    def finish(self):
        if hasattr(self, "_out"):
            self._out.release()



from typing import Type, Tuple

from wai.annotations.core.component import Component
from wai.annotations.core.domain import DomainSpecifier
from wai.annotations.core.specifier import SourceStageSpecifier


class VideoFileODInputFormatSpecifier(SourceStageSpecifier):
    """
    Base specifier for the label-dist in each known domain.
    """
    @classmethod
    def description(cls) -> str:
        return "Reads frames from a video file."

    @classmethod
    def components(cls) -> Tuple[Type[Component], ...]:
        from wai.annotations.video.format.video_file.component import VideoFileReader
        return VideoFileReader,

    """
    Specifier for from-video-file-od in the object-detection domain.
    """
    @classmethod
    def domain(cls) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.object_detection import ImageObjectDetectionDomainSpecifier
        return ImageObjectDetectionDomainSpecifier



from wai.common.cli.options import TypedOption
from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import RequiresNoFinalisation
from wai.annotations.domain.image import ImageInstance


class DropFrames(
    RequiresNoFinalisation,
    ProcessorComponent[ImageInstance, ImageInstance]
):
    """
    Stream processor which drops frames.
    """

    nth_frame: int = TypedOption(
        "-n", "--nth-frame",
        type=int,
        default=0,
        help="which nth frame to drop, e..g, '2' means to drop every 2nd frame; passes frames through if <=1"
    )

    def process_element(
            self,
            element: ImageInstance,
            then: ThenFunction[ImageInstance],
            done: DoneFunction
    ):
        # nothing to do?
        if self.nth_frame <= 1:
            then(element)
            return

        if not hasattr(self, "_count"):
            self._count = 0

        self._count += 1
        if (self._count % self.nth_frame) != 0:
            then(element)
        else:
            self._count = 0



from typing import Type, Tuple

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.domain import DomainSpecifier
from wai.annotations.core.specifier import ProcessorStageSpecifier


class SkipSimilarFramesISPSpecifier(ProcessorStageSpecifier):
    """
    Specifies the skip-similar-frames ISP.
    """
    @classmethod
    def description(cls) -> str:
        return "Skips frames in the stream that are deemed too similar."

    @classmethod
    def domain_transfer_function(
            cls,
            input_domain: Type[DomainSpecifier]
    ) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.classification import ImageClassificationDomainSpecifier
        from wai.annotations.domain.image.object_detection import ImageObjectDetectionDomainSpecifier
        from wai.annotations.domain.image.segmentation import ImageSegmentationDomainSpecifier
        if input_domain is ImageClassificationDomainSpecifier:
            return input_domain
        elif input_domain is ImageObjectDetectionDomainSpecifier:
            return input_domain
        elif input_domain is ImageSegmentationDomainSpecifier:
            return input_domain
        else:
            raise Exception(
                f"Crop only handles the following domains: "
                f"{ImageClassificationDomainSpecifier.name()}, "
                f"{ImageObjectDetectionDomainSpecifier.name()}, "
                f"{ImageSegmentationDomainSpecifier.name()}"
            )

    @classmethod
    def components(cls) -> Tuple[Type[ProcessorComponent]]:
        from wai.annotations.video.isp.skip_similar_frames.component import SkipSimilarFrames
        return SkipSimilarFrames,



import csv
import cv2
import io
import json
import numpy as np
import sys
import termplotlib as tpl

from wai.annotations.core.component import SinkComponent
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.common.cli.options import TypedOption, FlagOption
from wai.annotations.video.util.change_detection import detect_change


class CalcFrameChanges(
    SinkComponent[ImageObjectDetectionInstance]
):

    conversion: str = TypedOption(
        "-c", "--conversion",
        type=str,
        default="gray",
        help="how to convert the BGR image to a single channel image (gray/r/g/b)"
    )

    bw_threshold: int = TypedOption(
        "-b", "--bw-threshold",
        type=int,
        default=128,
        help="the threshold to use for converting a gray-scale like image to black and white (0-255)"
    )

    change_threshold: float = TypedOption(
        "-t", "--change-threshold",
        type=float,
        default=0.01,
        help="the percentage of pixels that changed relative to size of image (0-1)"
    )

    num_bins: int = TypedOption(
        "-B", "--num-bins",
        type=int,
        default=20,
        help="the number of bins to use for the histogram"
    )

    output_format: str = TypedOption(
        "-f", "--output-format",
        type=str,
        default="text",
        help="how to output the statistics (text/csv/json)"
    )

    output_file: str = TypedOption(
        "-o", "--output",
        type=str,
        default="",
        help="the file to write to statistics to, stdout if not provided"
    )

    verbose: bool = FlagOption(
        "-v", "--verbose",
        help="whether to output some debugging output."
    )

    def output_stats(self):
        """
        Calculates and outputs the statistics.
        """
        if not hasattr(self, "_ratios"):
            self.logger.error("Not data collected for statistics!")
            return

        use_stdout = len(self.output_file) == 0
        counts, bin_edges = np.histogram(self._ratios, bins=self.num_bins)

        # text
        if self.output_format == "text":
            fig = tpl.figure()
            fig.hist(counts, bin_edges, orientation="horizontal", force_ascii=False)
            if use_stdout:
                fig.show()
            else:
                with open(self.output_file, "w") as fp:
                    fp.write(fig.get_string())
                    fp.write("\n")

        # csv
        elif self.output_format == "csv":
            data = [["bin", "from", "to", "count"]]
            for i in range(self.num_bins):
                data.append([i, bin_edges[i], bin_edges[i+1], counts[i]])
            if use_stdout:
                writer = csv.writer(sys.stdout)
                writer.writerows(data)
            else:
                with open(self.output_file, "w") as fp:
                    writer = csv.writer(fp)
                    writer.writerows(data)

        # json
        elif self.output_format == "json":
            data = []
            for i in range(self.num_bins):
                data.append({
                    "bin": i,
                    "from": float(bin_edges[i]),
                    "to": float(bin_edges[i+1]),
                    "count": int(counts[i])
                })
            if use_stdout:
                print(json.dumps(data, indent=2))
            else:
                with open(self.output_file, "w") as fp:
                    json.dump(data, fp, indent=2)

    def consume_element(self, element: ImageObjectDetectionInstance):
        """
        Consumes instances.
        """
        if not hasattr(self, "_last_image"):
            self._last_image = None
            self._ratios = []

        # read image
        img_array = np.fromstring(io.BytesIO(element.data.data).read(), dtype=np.uint8)
        img = cv2.imdecode(img_array, cv2.IMREAD_COLOR)

        # nothing to compare against?
        if self._last_image is None:
            # shift state
            self._last_image = img
            return

        # detect change
        ratio, changed = detect_change(self._last_image, img,
                                       self.conversion, self.bw_threshold, self.change_threshold)
        if self.verbose:
            self.logger.info("%s (ratio/changed): %f -> %s" % (element.data.filename, ratio, str(changed)))

        if changed:
            # shift state
            self._last_image = img
            self._ratios.append(ratio)

    def finish(self):
        self.output_stats()



import cv2
import os

from wai.annotations.core.component import SourceComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.domain.image import Image, ImageFormat
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance

from wai.common.cli.options import TypedOption
from wai.common.adams.imaging.locateobjects import LocatedObjects


class VideoFileReader(
    SourceComponent[ImageObjectDetectionInstance]
):

    input_file: str = TypedOption(
        "-i", "--input",
        type=str,
        default="",
        help="the video file to read"
    )

    from_frame: int = TypedOption(
        "-f", "--from-frame",
        type=int,
        default=1,
        help="determines with which frame to start the stream (1-based index)"
    )

    to_frame: int = TypedOption(
        "-t", "--to-frame",
        type=int,
        default=-1,
        help="determines after which frame to stop (1-based index); ignored if <=0"
    )

    nth_frame: int = TypedOption(
        "-n", "--nth-frame",
        type=int,
        default=1,
        help="determines whether frames get skipped and only evert nth frame gets forwarded"
    )

    max_frames: int = TypedOption(
        "-m", "--max-frames",
        type=int,
        default=-1,
        help="determines the maximum number of frames to read; ignored if <=0"
    )

    prefix: str = TypedOption(
        "-p", "--prefix",
        type=str,
        default="",
        help="the prefix to use for the frames"
    )

    """
    The source of elements in a stream.
    """
    def produce(
            self,
            then: ThenFunction[ImageObjectDetectionInstance],
            done: DoneFunction
    ):
        """
        Produces elements and inserts them into the stream. Should call 'then'
        for each element produced, and then call 'done' when finished.

        :param then:    A function which forwards elements into the stream.
        :param done:    A function which closes the stream when called.
        """
        # open video file
        if not hasattr(self, "_cap"):
            self._cap = cv2.VideoCapture(self.input_file)
            self._frame_no = 0
            self._frame_count = 0

        # next frame?
        count = 0
        while (self._cap is not None) and self._cap.isOpened():
            # next frame
            self._frame_no += 1
            count += 1
            retval, frame_curr = self._cap.read()

            if retval:
                # within frame window?
                if self.from_frame > 0:
                    if self._frame_no < self.from_frame:
                        continue
                if self.to_frame > 0:
                    if self._frame_no >= self.to_frame:
                        break

                # skip frame?
                if (self.nth_frame > 1) and (count < self.nth_frame):
                    continue

                # max frames reached?
                if (self.max_frames > 0) and (self._frame_count >= self.max_frames):
                    break

                self._frame_count += 1
                count = 0
                data = cv2.imencode(".jpg", frame_curr)[1].tobytes()
                prefix = (os.path.splitext(os.path.basename(self.input_file))[0] + "-") if (len(self.prefix) == 0) else self.prefix
                filename = os.path.join(
                    os.path.dirname(self.input_file),
                    "%s%08d.jpg" % (prefix, self._frame_no))
                height, width, _ = frame_curr.shape
                image = Image(filename=filename, data=data, format=ImageFormat.JPG, size=(width, height))
                instance = ImageObjectDetectionInstance(data=image, annotations=LocatedObjects())
                then(instance)
            else:
                self._cap.release()
                self._cap = None
                done()

        # close video file
        if self._cap is not None:
            self._cap.release()
            done()



import cv2
import io
import numpy as np

from wai.common.cli.options import TypedOption, FlagOption
from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import RequiresNoFinalisation
from wai.annotations.domain.image import ImageInstance
from wai.annotations.video.util.change_detection import detect_change


class SkipSimilarFrames(
    RequiresNoFinalisation,
    ProcessorComponent[ImageInstance, ImageInstance]
):
    """
    Stream processor which skips similar frames.
    """

    conversion: str = TypedOption(
        "-c", "--conversion",
        type=str,
        default="gray",
        help="how to convert the BGR image to a single channel image (gray/r/g/b)"
    )

    bw_threshold: int = TypedOption(
        "-b", "--bw-threshold",
        type=int,
        default=128,
        help="the threshold to use for converting a gray-scale like image to black and white (0-255)"
    )

    change_threshold: float = TypedOption(
        "-t", "--change-threshold",
        type=float,
        default=0.01,
        help="the percentage of pixels that changed relative to size of image (0-1)"
    )

    verbose: bool = FlagOption(
        "-v", "--verbose",
        help="whether to output some debugging output."
    )

    def process_element(
            self,
            element: ImageInstance,
            then: ThenFunction[ImageInstance],
            done: DoneFunction
    ):
        if not hasattr(self, "_last_image"):
            self._last_image = None

        # read image
        img_array = np.fromstring(io.BytesIO(element.data.data).read(), dtype=np.uint8)
        img = cv2.imdecode(img_array, cv2.IMREAD_COLOR)

        # nothing to compare against?
        if self._last_image is None:
            # shift state
            self._last_image = img
            # forward frame
            then(element)
            return

        # detect change
        ratio, changed = detect_change(self._last_image, img,
                                       self.conversion, self.bw_threshold, self.change_threshold)
        if self.verbose:
            self.logger.info("%s (ratio/changed): %f -> %s" % (element.data.filename, ratio, str(changed)))

        if changed:
            # shift state
            self._last_image = img
            # forward frame
            then(element)
